import itertools
from pprint import pprint
from qiskit import QuantumCircuit, Aer, execute
from qiskit.quantum_info import Statevector
backend = Aer.get_backend('statevector_simulator')  
import numpy as np

maxgc = 4
gs = ['x','ccx']
gs_opcode = ['0','1']
qbnos = 4
qb = list(range(0,qbnos))

def toStr(n,base):
    '''
    Convert a decimal number to base-n
    '''
    convertString = "0123456789ABCDEF"
    if n < base:
        return convertString[n]
    else:
        return toStr(n//base,base) + convertString[n%base]

def X_choice():
    ''' 
    Creates combination of X opcode and 1 qubits 
    '''
    c = list(itertools.combinations(qb, 1))
    allc = []
    for i in c:
        allc.append(gs_opcode[0]+str(i[0]))
    return allc

def CCX_choice():
    ''' 
    Creates combination of CCX opcode and 3 qubits (with 2 controls indistinguisable) 
    '''
    c = list(itertools.combinations(qb, 3))
    allc = []
    for i in c:
        allc.append(gs_opcode[1]+str(i[0])+str(i[1])+str(i[2]))
        allc.append(gs_opcode[1]+str(i[1])+str(i[2])+str(i[0]))
        allc.append(gs_opcode[1]+str(i[2])+str(i[0])+str(i[1]))
    return allc

def progs(gc):
    ''' 
    Generates all programs with gate count = gc 
    '''
    p = []
    for i in range(0,len(gs)**gc):
        gseq = toStr(i,len(gs)).zfill(gc)
        cg = ['']          
        for j in range(0, len(gseq)):
            if gseq[j] == '0':
                g = list(itertools.product(cg, setX))
            if gseq[j] == '1':
                g = list(itertools.product(cg, setCCX))
            cg = []
            for k in g:
                cg.append(''.join(map(str, k)))
        for j in cg:
            p.append(j)
    return p

possible_states = []
for i in range(0,2**qbnos):
    possible_states.append(format(i, '04b'))    

def state_construction(index,qc): # How will this change for H, CCX ??
    '''
    For considering different initial states
    '''
    for i in range(4):
        if(index[-i-1]) == '1':
            qc.x(i)

def runQprog(desc):
    
    for init in range(0,len(possible_states)):
        qcirc = QuantumCircuit(qbnos)
        state_construction(possible_states[init],qcirc)
        qcirc.barrier()
        i = 0
        while (i < len(desc)):
            if desc[i]=='0':
                qcirc.x(int(desc[i+1]))
                i+= 2
            elif desc[i]=='1':
                qcirc.ccx(int(desc[i+1]),int(desc[i+2]),int(desc[i+3]))
                i+= 4
        qcirc.barrier()
        
        job = execute(qcirc, backend, shots=1, memory=True)
        result = job.result()
        memory = Statevector(result.get_statevector(qcirc)).probabilities()
        for j in range(0,len(memory)):
            if memory[j] > 0:
                AP[init][j] += memory[j]
    return AP

setX = X_choice()
setCCX = CCX_choice()

np.set_printoptions(precision=4)

gc = 1
p = progs(gc)
AP = np.zeros((16,16))
for i in p:
    runQprog(i)
M1 = np.divide(AP,len(p))
print("nGC Level:",gc)
print(M1)

def AP_next(gc):
    Mi = M1
    for i in range(1,gc):
        Mi = Mi@M1
    return Mi

gc = 2
p = progs(gc)
AP = np.zeros((16,16))
for i in p:
    runQprog(i)
M2 = np.divide(AP,len(p))
print("nGC Level:",gc)
print(M2)

M2_bao = AP_next(2)
print("nGC Level:",gc)
print(M2_bao)
print("nGC Level",gc,"error")
print(M2_bao-M2)

gc = 3
p = progs(gc)
AP = np.zeros((16,16))
for i in p:
    runQprog(i)
M3 = np.divide(AP,len(p))
print("nGC Level:",gc)
print(M3)

M3_bao = AP_next(3)
print("nGC Level:",gc)
print(M3_bao)
print("nGC Level",gc,"error")
print(M3_bao-M3)

# Output Log

'''

(qeait) D:\GoogleDrive\11 - RESEARCH\A1 - Programs\AIG>python c_AP.py

nGC Level: 1
[[0.75   0.0625 0.0625 0.     0.0625 0.     0.     0.     0.0625 0.     0.     0.     0.     0.     0.     0.    ]
 [0.0625 0.75   0.     0.0625 0.     0.0625 0.     0.     0.     0.0625 0.     0.     0.     0.     0.     0.    ]
 [0.0625 0.     0.75   0.0625 0.     0.     0.0625 0.     0.     0.     0.0625 0.     0.     0.     0.     0.    ]
 [0.     0.0625 0.0625 0.625  0.     0.     0.     0.125  0.     0.     0.     0.125  0.     0.     0.     0.    ]
 [0.0625 0.     0.     0.     0.75   0.0625 0.0625 0.     0.     0.     0.     0.     0.0625 0.     0.     0.    ]
 [0.     0.0625 0.     0.     0.0625 0.625  0.     0.125  0.     0.     0.     0.     0.     0.125  0.     0.    ]
 [0.     0.     0.0625 0.     0.0625 0.     0.625  0.125  0.     0.     0.     0.     0.     0.     0.125  0.    ]
 [0.     0.     0.     0.125  0.     0.125  0.125  0.375  0.     0.     0.     0.     0.     0.     0.     0.25  ]
 [0.0625 0.     0.     0.     0.     0.     0.     0.     0.75   0.0625 0.0625 0.     0.0625 0.     0.     0.    ]
 [0.     0.0625 0.     0.     0.     0.     0.     0.     0.0625 0.625  0.     0.125  0.     0.125  0.     0.    ]
 [0.     0.     0.0625 0.     0.     0.     0.     0.     0.0625 0.     0.625  0.125  0.     0.     0.125  0.    ]
 [0.     0.     0.     0.125  0.     0.     0.     0.     0.     0.125  0.125  0.375  0.     0.     0.     0.25  ]
 [0.     0.     0.     0.     0.0625 0.     0.     0.     0.0625 0.     0.     0.     0.625  0.125  0.125  0.    ]
 [0.     0.     0.     0.     0.     0.125  0.     0.     0.     0.125  0.     0.     0.125  0.375  0.     0.25  ]
 [0.     0.     0.     0.     0.     0.     0.125  0.     0.     0.     0.125  0.     0.125  0.     0.375  0.25  ]
 [0.     0.     0.     0.     0.     0.     0.     0.25   0.     0.     0.     0.25   0.     0.25   0.25   0.    ]]

nGC Level: 2
[[0.5781 0.0938 0.0938 0.0078 0.0938 0.0078 0.0078 0.     0.0938 0.0078 0.0078 0.     0.0078 0.     0.     0.    ]
 [0.0938 0.5781 0.0078 0.0859 0.0078 0.0859 0.     0.0156 0.0078 0.0859 0.     0.0156 0.     0.0156 0.     0.    ]
 [0.0938 0.0078 0.5781 0.0859 0.0078 0.     0.0859 0.0156 0.0078 0.     0.0859 0.0156 0.     0.     0.0156 0.    ]
 [0.0078 0.0859 0.0859 0.4297 0.     0.0195 0.0195 0.125  0.     0.0195 0.0195 0.125  0.     0.     0.     0.0625]
 [0.0938 0.0078 0.0078 0.     0.5781 0.0859 0.0859 0.0156 0.0078 0.     0.     0.     0.0859 0.0156 0.0156 0.    ]
 [0.0078 0.0859 0.     0.0195 0.0859 0.4297 0.0195 0.125  0.     0.0195 0.     0.     0.0195 0.125  0.     0.0625]
 [0.0078 0.     0.0859 0.0195 0.0859 0.0195 0.4297 0.125  0.     0.     0.0195 0.     0.0195 0.     0.125  0.0625]
 [0.     0.0156 0.0156 0.125  0.0156 0.125  0.125  0.25   0.     0.     0.     0.0781 0.     0.0781 0.0781 0.0938]
 [0.0938 0.0078 0.0078 0.     0.0078 0.     0.     0.     0.5781 0.0859 0.0859 0.0156 0.0859 0.0156 0.0156 0.    ]
 [0.0078 0.0859 0.     0.0195 0.     0.0195 0.     0.     0.0859 0.4297 0.0195 0.125  0.0195 0.125  0.     0.0625]
 [0.0078 0.     0.0859 0.0195 0.     0.     0.0195 0.     0.0859 0.0195 0.4297 0.125  0.0195 0.     0.125  0.0625]
 [0.     0.0156 0.0156 0.125  0.     0.     0.     0.0781 0.0156 0.125  0.125  0.25   0.     0.0781 0.0781 0.0938]
 [0.0078 0.     0.     0.     0.0859 0.0195 0.0195 0.     0.0859 0.0195 0.0195 0.     0.4297 0.125  0.125  0.0625]
 [0.     0.0156 0.     0.     0.0156 0.125  0.     0.0781 0.0156 0.125  0.     0.0781 0.125  0.25   0.0781 0.0938]
 [0.     0.     0.0156 0.     0.0156 0.     0.125  0.0781 0.0156 0.     0.125  0.0781 0.125  0.0781 0.25   0.0938]
 [0.     0.     0.     0.0625 0.     0.0625 0.0625 0.0938 0.     0.0625 0.0625 0.0938 0.0625 0.0938 0.0938 0.25  ]]

nGC Level: 2
[[0.5781 0.0938 0.0938 0.0078 0.0938 0.0078 0.0078 0.     0.0938 0.0078 0.0078 0.     0.0078 0.     0.     0.    ]
 [0.0938 0.5781 0.0078 0.0859 0.0078 0.0859 0.     0.0156 0.0078 0.0859 0.     0.0156 0.     0.0156 0.     0.    ]
 [0.0938 0.0078 0.5781 0.0859 0.0078 0.     0.0859 0.0156 0.0078 0.     0.0859 0.0156 0.     0.     0.0156 0.    ]
 [0.0078 0.0859 0.0859 0.4297 0.     0.0195 0.0195 0.125  0.     0.0195 0.0195 0.125  0.     0.     0.     0.0625]
 [0.0938 0.0078 0.0078 0.     0.5781 0.0859 0.0859 0.0156 0.0078 0.     0.     0.     0.0859 0.0156 0.0156 0.    ]
 [0.0078 0.0859 0.     0.0195 0.0859 0.4297 0.0195 0.125  0.     0.0195 0.     0.     0.0195 0.125  0.     0.0625]
 [0.0078 0.     0.0859 0.0195 0.0859 0.0195 0.4297 0.125  0.     0.     0.0195 0.     0.0195 0.     0.125  0.0625]
 [0.     0.0156 0.0156 0.125  0.0156 0.125  0.125  0.25   0.     0.     0.     0.0781 0.     0.0781 0.0781 0.0938]
 [0.0938 0.0078 0.0078 0.     0.0078 0.     0.     0.     0.5781 0.0859 0.0859 0.0156 0.0859 0.0156 0.0156 0.    ]
 [0.0078 0.0859 0.     0.0195 0.     0.0195 0.     0.     0.0859 0.4297 0.0195 0.125  0.0195 0.125  0.     0.0625]
 [0.0078 0.     0.0859 0.0195 0.     0.     0.0195 0.     0.0859 0.0195 0.4297 0.125  0.0195 0.     0.125  0.0625]
 [0.     0.0156 0.0156 0.125  0.     0.     0.     0.0781 0.0156 0.125  0.125  0.25   0.     0.0781 0.0781 0.0938]
 [0.0078 0.     0.     0.     0.0859 0.0195 0.0195 0.     0.0859 0.0195 0.0195 0.     0.4297 0.125  0.125  0.0625]
 [0.     0.0156 0.     0.     0.0156 0.125  0.     0.0781 0.0156 0.125  0.     0.0781 0.125  0.25   0.0781 0.0938]
 [0.     0.     0.0156 0.     0.0156 0.     0.125  0.0781 0.0156 0.     0.125  0.0781 0.125  0.0781 0.25   0.0938]
 [0.     0.     0.     0.0625 0.     0.0625 0.0625 0.0938 0.     0.0625 0.0625 0.0938 0.0625 0.0938 0.0938 0.25  ]]

nGC Level 2 error
[[0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]]

nGC Level: 3
[[0.457  0.1079 0.1079 0.0166 0.1079 0.0166 0.0166 0.0029 0.1079 0.0166 0.0166 0.0029 0.0166 0.0029 0.0029 0.    ]
 [0.1079 0.4556 0.0171 0.0942 0.0171 0.0942 0.0029 0.0273 0.0171 0.0942 0.0029 0.0273 0.0029 0.0273 0.     0.0117]
 [0.1079 0.0171 0.4556 0.0942 0.0171 0.0029 0.0942 0.0273 0.0171 0.0029 0.0942 0.0273 0.0029 0.     0.0273 0.0117]
 [0.0166 0.0942 0.0942 0.3105 0.0029 0.0332 0.0332 0.1211 0.0029 0.0332 0.0332 0.1211 0.     0.0205 0.0205 0.0625]
 [0.1079 0.0171 0.0171 0.0029 0.4556 0.0942 0.0942 0.0273 0.0171 0.0029 0.0029 0.     0.0942 0.0273 0.0273 0.0117]
 [0.0166 0.0942 0.0029 0.0332 0.0942 0.3105 0.0332 0.1211 0.0029 0.0332 0.     0.0205 0.0332 0.1211 0.0205 0.0625]
 [0.0166 0.0029 0.0942 0.0332 0.0942 0.0332 0.3105 0.1211 0.0029 0.     0.0332 0.0205 0.0332 0.0205 0.1211 0.0625]
 [0.0029 0.0273 0.0273 0.1211 0.0273 0.1211 0.1211 0.1641 0.     0.0205 0.0205 0.0684 0.0205 0.0684 0.0684 0.1211]
 [0.1079 0.0171 0.0171 0.0029 0.0171 0.0029 0.0029 0.     0.4556 0.0942 0.0942 0.0273 0.0942 0.0273 0.0273 0.0117]
 [0.0166 0.0942 0.0029 0.0332 0.0029 0.0332 0.     0.0205 0.0942 0.3105 0.0332 0.1211 0.0332 0.1211 0.0205 0.0625]
 [0.0166 0.0029 0.0942 0.0332 0.0029 0.     0.0332 0.0205 0.0942 0.0332 0.3105 0.1211 0.0332 0.0205 0.1211 0.0625]
 [0.0029 0.0273 0.0273 0.1211 0.     0.0205 0.0205 0.0684 0.0273 0.1211 0.1211 0.1641 0.0205 0.0684 0.0684 0.1211]
 [0.0166 0.0029 0.0029 0.     0.0942 0.0332 0.0332 0.0205 0.0942 0.0332 0.0332 0.0205 0.3105 0.1211 0.1211 0.0625]
 [0.0029 0.0273 0.     0.0205 0.0273 0.1211 0.0205 0.0684 0.0273 0.1211 0.0205 0.0684 0.1211 0.1641 0.0684 0.1211]
 [0.0029 0.     0.0273 0.0205 0.0273 0.0205 0.1211 0.0684 0.0273 0.0205 0.1211 0.0684 0.1211 0.0684 0.1641 0.1211]
 [0.     0.0117 0.0117 0.0625 0.0117 0.0625 0.0625 0.1211 0.0117 0.0625 0.0625 0.1211 0.0625 0.1211 0.1211 0.0938]]

nGC Level: 3
[[0.457  0.1079 0.1079 0.0166 0.1079 0.0166 0.0166 0.0029 0.1079 0.0166 0.0166 0.0029 0.0166 0.0029 0.0029 0.    ]
 [0.1079 0.4556 0.0171 0.0942 0.0171 0.0942 0.0029 0.0273 0.0171 0.0942 0.0029 0.0273 0.0029 0.0273 0.     0.0117]
 [0.1079 0.0171 0.4556 0.0942 0.0171 0.0029 0.0942 0.0273 0.0171 0.0029 0.0942 0.0273 0.0029 0.     0.0273 0.0117]
 [0.0166 0.0942 0.0942 0.3105 0.0029 0.0332 0.0332 0.1211 0.0029 0.0332 0.0332 0.1211 0.     0.0205 0.0205 0.0625]
 [0.1079 0.0171 0.0171 0.0029 0.4556 0.0942 0.0942 0.0273 0.0171 0.0029 0.0029 0.     0.0942 0.0273 0.0273 0.0117]
 [0.0166 0.0942 0.0029 0.0332 0.0942 0.3105 0.0332 0.1211 0.0029 0.0332 0.     0.0205 0.0332 0.1211 0.0205 0.0625]
 [0.0166 0.0029 0.0942 0.0332 0.0942 0.0332 0.3105 0.1211 0.0029 0.     0.0332 0.0205 0.0332 0.0205 0.1211 0.0625]
 [0.0029 0.0273 0.0273 0.1211 0.0273 0.1211 0.1211 0.1641 0.     0.0205 0.0205 0.0684 0.0205 0.0684 0.0684 0.1211]
 [0.1079 0.0171 0.0171 0.0029 0.0171 0.0029 0.0029 0.     0.4556 0.0942 0.0942 0.0273 0.0942 0.0273 0.0273 0.0117]
 [0.0166 0.0942 0.0029 0.0332 0.0029 0.0332 0.     0.0205 0.0942 0.3105 0.0332 0.1211 0.0332 0.1211 0.0205 0.0625]
 [0.0166 0.0029 0.0942 0.0332 0.0029 0.     0.0332 0.0205 0.0942 0.0332 0.3105 0.1211 0.0332 0.0205 0.1211 0.0625]
 [0.0029 0.0273 0.0273 0.1211 0.     0.0205 0.0205 0.0684 0.0273 0.1211 0.1211 0.1641 0.0205 0.0684 0.0684 0.1211]
 [0.0166 0.0029 0.0029 0.     0.0942 0.0332 0.0332 0.0205 0.0942 0.0332 0.0332 0.0205 0.3105 0.1211 0.1211 0.0625]
 [0.0029 0.0273 0.     0.0205 0.0273 0.1211 0.0205 0.0684 0.0273 0.1211 0.0205 0.0684 0.1211 0.1641 0.0684 0.1211]
 [0.0029 0.     0.0273 0.0205 0.0273 0.0205 0.1211 0.0684 0.0273 0.0205 0.1211 0.0684 0.1211 0.0684 0.1641 0.1211]
 [0.     0.0117 0.0117 0.0625 0.0117 0.0625 0.0625 0.1211 0.0117 0.0625 0.0625 0.1211 0.0625 0.1211 0.1211 0.0938]]

nGC Level 3 error
[[0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]
 [0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]]
 
'''