import itertools
from pprint import pprint
from qiskit import QuantumCircuit, Aer, execute
from qiskit.quantum_info import Statevector
backend = Aer.get_backend('statevector_simulator')  
import numpy as np

maxgc = 4
gs = ['h','ccx']
gs_opcode = ['0','1']
qbnos = 4
qb = list(range(0,qbnos))

def toStr(n,base):
    '''
    Convert a decimal number to base-n
    '''
    convertString = "0123456789ABCDEF"
    if n < base:
        return convertString[n]
    else:
        return toStr(n//base,base) + convertString[n%base]

def H_choice():
    ''' 
    Creates combination of H opcode and 1 qubits 
    '''
    c = list(itertools.combinations(qb, 1))
    allc = []
    for i in c:
        allc.append(gs_opcode[0]+str(i[0]))
    return allc

def CCX_choice():
    ''' 
    Creates combination of CCX opcode and 3 qubits (with 2 controls indistinguisable) 
    '''
    c = list(itertools.combinations(qb, 3))
    allc = []
    for i in c:
        allc.append(gs_opcode[1]+str(i[0])+str(i[1])+str(i[2]))
        allc.append(gs_opcode[1]+str(i[1])+str(i[2])+str(i[0]))
        allc.append(gs_opcode[1]+str(i[2])+str(i[0])+str(i[1]))
    return allc

def progs(gc):
    ''' 
    Generates all programs with gate count = gc 
    '''
    p = []
    for i in range(0,len(gs)**gc):
        gseq = toStr(i,len(gs)).zfill(gc)
        cg = ['']          
        for j in range(0, len(gseq)):
            if gseq[j] == '0':
                g = list(itertools.product(cg, setH))
            if gseq[j] == '1':
                g = list(itertools.product(cg, setCCX))
            cg = []
            for k in g:
                cg.append(''.join(map(str, k)))
        for j in cg:
            p.append(j)
    return p

possible_states = []
for i in range(0,2**qbnos):
    possible_states.append(format(i, '04b'))    

def state_construction(index,qc): # How will this change for H, CCX ??
    '''
    For considering different initial states
    '''
    for i in range(4):
        # qc.h(i)
        if(index[-i-1]) == '1':
            qc.x(i)
            # qc.h(i)

def runQprog(desc):
    
    for init in range(0,len(possible_states)):
        qcirc = QuantumCircuit(qbnos)
        state_construction(possible_states[init],qcirc)
        qcirc.barrier()
        i = 0
        while (i < len(desc)):
            if desc[i]=='0':
                qcirc.h(int(desc[i+1]))
                i+= 2
            elif desc[i]=='1':
                qcirc.ccx(int(desc[i+1]),int(desc[i+2]),int(desc[i+3]))
                i+= 4
        qcirc.barrier()
        
        job = execute(qcirc, backend, shots=1, memory=True)
        result = job.result()
        memory = Statevector(result.get_statevector(qcirc)).probabilities()
        for j in range(0,len(memory)):
            if memory[j] > 0:
                AP[init][j] += memory[j]
    return AP

setH = H_choice()
setCCX = CCX_choice()

np.set_printoptions(precision=4)

gc = 1
p = progs(gc)
AP = np.zeros((16,16))
for i in p:
    runQprog(i)
M1 = np.divide(AP,len(p))
print("nGC Level:",gc)
print(M1)

def AP_next(gc):
    Mi = M1
    for i in range(1,gc):
        Mi = Mi@M1
    return Mi

gc = 2
p = progs(gc)
AP = np.zeros((16,16))
for i in p:
    runQprog(i)
M2 = np.divide(AP,len(p))
print("nGC Level:",gc)
print(M2)

M2_bao = AP_next(2)
print("nGC Level:",gc)
print(M2_bao)
print("nGC Level",gc,"error")
print(np.round(M2_bao-M2,2))

gc = 3
p = progs(gc)
AP = np.zeros((16,16))
for i in p:
    runQprog(i)
M3 = np.divide(AP,len(p))
print("nGC Level:",gc)
print(M3)

M3_bao = AP_next(3)
print("nGC Level:",gc)
print(M3_bao)
print("nGC Level",gc,"error")
print(np.round(M3_bao-M3,2))

# Output Log

'''

(qeait) D:\GoogleDrive\11 - RESEARCH\A1 - Programs\AIG>python q_AP.py

nGC Level: 1
[[0.875  0.0313 0.0313 0.     0.0313 0.     0.     0.     0.0313 0.     0.     0.     0.     0.     0.     0.    ]
 [0.0313 0.875  0.     0.0313 0.     0.0313 0.     0.     0.     0.0313 0.     0.     0.     0.     0.     0.    ]
 [0.0313 0.     0.875  0.0313 0.     0.     0.0313 0.     0.     0.     0.0313 0.     0.     0.     0.     0.    ]
 [0.     0.0313 0.0313 0.75   0.     0.     0.     0.0938 0.     0.     0.     0.0938 0.     0.     0.     0.    ]
 [0.0313 0.     0.     0.     0.875  0.0313 0.0313 0.     0.     0.     0.     0.     0.0313 0.     0.     0.    ]
 [0.     0.0313 0.     0.     0.0313 0.75   0.     0.0938 0.     0.     0.     0.     0.     0.0938 0.     0.    ]
 [0.     0.     0.0313 0.     0.0313 0.     0.75   0.0938 0.     0.     0.     0.     0.     0.     0.0938 0.    ]
 [0.     0.     0.     0.0938 0.     0.0938 0.0938 0.5    0.     0.     0.     0.     0.     0.     0.     0.2188]
 [0.0313 0.     0.     0.     0.     0.     0.     0.     0.875  0.0313 0.0313 0.     0.0313 0.     0.     0.    ]
 [0.     0.0313 0.     0.     0.     0.     0.     0.     0.0313 0.75   0.     0.0938 0.     0.0938 0.     0.    ]
 [0.     0.     0.0313 0.     0.     0.     0.     0.     0.0313 0.     0.75   0.0938 0.     0.     0.0938 0.    ]
 [0.     0.     0.     0.0938 0.     0.     0.     0.     0.     0.0938 0.0938 0.5    0.     0.     0.     0.2188]
 [0.     0.     0.     0.     0.0313 0.     0.     0.     0.0313 0.     0.     0.     0.75   0.0938 0.0938 0.    ]
 [0.     0.     0.     0.     0.     0.0938 0.     0.     0.     0.0938 0.     0.     0.0938 0.5    0.     0.2188]
 [0.     0.     0.     0.     0.     0.     0.0938 0.     0.     0.     0.0938 0.     0.0938 0.     0.5    0.2188]
 [0.     0.     0.     0.     0.     0.     0.     0.2188 0.     0.     0.     0.2188 0.     0.2188 0.2188 0.125 ]]

nGC Level: 2
[[0.7773 0.0527 0.0527 0.002  0.0527 0.002  0.002  0.     0.0527 0.002  0.002  0.     0.002  0.     0.     0.    ]
 [0.0527 0.7773 0.002  0.0488 0.002  0.0488 0.     0.0059 0.002  0.0488 0.     0.0059 0.     0.0059 0.     0.    ]
 [0.0527 0.002  0.7773 0.0488 0.002  0.     0.0488 0.0059 0.002  0.     0.0488 0.0059 0.     0.     0.0059 0.    ]
 [0.002  0.0488 0.0488 0.5898 0.     0.0098 0.0098 0.1152 0.     0.0098 0.0098 0.1152 0.     0.     0.     0.041 ]
 [0.0527 0.002  0.002  0.     0.7773 0.0488 0.0488 0.0059 0.002  0.     0.     0.     0.0488 0.0059 0.0059 0.    ]
 [0.002  0.0488 0.     0.0098 0.0488 0.5898 0.0098 0.1152 0.     0.0098 0.     0.     0.0098 0.1152 0.     0.041 ]
 [0.002  0.     0.0488 0.0098 0.0488 0.0098 0.5898 0.1152 0.     0.     0.0098 0.     0.0098 0.     0.1152 0.041 ]
 [0.     0.0059 0.0059 0.1152 0.0059 0.1152 0.1152 0.332  0.     0.     0.     0.0566 0.     0.0566 0.0566 0.1348]
 [0.0527 0.002  0.002  0.     0.002  0.     0.     0.     0.7773 0.0488 0.0488 0.0059 0.0488 0.0059 0.0059 0.    ]
 [0.002  0.0488 0.     0.0098 0.     0.0098 0.     0.     0.0488 0.5898 0.0098 0.1152 0.0098 0.1152 0.     0.041 ]
 [0.002  0.     0.0488 0.0098 0.     0.     0.0098 0.     0.0488 0.0098 0.5898 0.1152 0.0098 0.     0.1152 0.041 ]
 [0.     0.0059 0.0059 0.1152 0.     0.     0.     0.0566 0.0059 0.1152 0.1152 0.332  0.     0.0566 0.0566 0.1348]
 [0.002  0.     0.     0.     0.0488 0.0098 0.0098 0.     0.0488 0.0098 0.0098 0.     0.5898 0.1152 0.1152 0.041 ]
 [0.     0.0059 0.     0.     0.0059 0.1152 0.     0.0566 0.0059 0.1152 0.     0.0566 0.1152 0.332  0.0566 0.1348]
 [0.     0.     0.0059 0.     0.0059 0.     0.1152 0.0566 0.0059 0.     0.1152 0.0566 0.1152 0.0566 0.332  0.1348]
 [0.     0.     0.     0.041  0.     0.041  0.041  0.1348 0.     0.041  0.041  0.1348 0.041  0.1348 0.1348 0.2148]]

nGC Level: 2
[[0.7695 0.0547 0.0547 0.002  0.0547 0.002  0.002  0.     0.0547 0.002  0.002  0.     0.002  0.     0.     0.    ]
 [0.0547 0.7695 0.002  0.0508 0.002  0.0508 0.     0.0059 0.002  0.0508 0.     0.0059 0.     0.0059 0.     0.    ]
 [0.0547 0.002  0.7695 0.0508 0.002  0.     0.0508 0.0059 0.002  0.     0.0508 0.0059 0.     0.     0.0059 0.    ]
 [0.002  0.0508 0.0508 0.582  0.     0.0098 0.0098 0.1172 0.     0.0098 0.0098 0.1172 0.     0.     0.     0.041 ]
 [0.0547 0.002  0.002  0.     0.7695 0.0508 0.0508 0.0059 0.002  0.     0.     0.     0.0508 0.0059 0.0059 0.    ]
 [0.002  0.0508 0.     0.0098 0.0508 0.582  0.0098 0.1172 0.     0.0098 0.     0.     0.0098 0.1172 0.     0.041 ]
 [0.002  0.     0.0508 0.0098 0.0508 0.0098 0.582  0.1172 0.     0.     0.0098 0.     0.0098 0.     0.1172 0.041 ]
 [0.     0.0059 0.0059 0.1172 0.0059 0.1172 0.1172 0.3242 0.     0.     0.     0.0566 0.     0.0566 0.0566 0.1367]
 [0.0547 0.002  0.002  0.     0.002  0.     0.     0.     0.7695 0.0508 0.0508 0.0059 0.0508 0.0059 0.0059 0.    ]
 [0.002  0.0508 0.     0.0098 0.     0.0098 0.     0.     0.0508 0.582  0.0098 0.1172 0.0098 0.1172 0.     0.041 ]
 [0.002  0.     0.0508 0.0098 0.     0.     0.0098 0.     0.0508 0.0098 0.582  0.1172 0.0098 0.     0.1172 0.041 ]
 [0.     0.0059 0.0059 0.1172 0.     0.     0.     0.0566 0.0059 0.1172 0.1172 0.3242 0.     0.0566 0.0566 0.1367]
 [0.002  0.     0.     0.     0.0508 0.0098 0.0098 0.     0.0508 0.0098 0.0098 0.     0.582  0.1172 0.1172 0.041 ]
 [0.     0.0059 0.     0.     0.0059 0.1172 0.     0.0566 0.0059 0.1172 0.     0.0566 0.1172 0.3242 0.0566 0.1367]
 [0.     0.     0.0059 0.     0.0059 0.     0.1172 0.0566 0.0059 0.     0.1172 0.0566 0.1172 0.0566 0.3242 0.1367]
 [0.     0.     0.     0.041  0.     0.041  0.041  0.1367 0.     0.041  0.041  0.1367 0.041  0.1367 0.1367 0.207 ]]

nGC Level 2 error
[[-0.01  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.   -0.01  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.   -0.01  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.   -0.01  0.   -0.   -0.    0.    0.   -0.   -0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.   -0.01  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.   -0.    0.   -0.01 -0.    0.    0.   -0.    0.    0.   -0.    0.    0.    0.  ]
 [ 0.    0.    0.   -0.    0.   -0.   -0.01  0.    0.    0.   -0.    0.   -0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.   -0.01  0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.   -0.01  0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.   -0.    0.   -0.    0.    0.    0.   -0.01 -0.    0.   -0.    0.    0.    0.  ]
 [ 0.    0.    0.   -0.    0.    0.   -0.    0.    0.   -0.   -0.01  0.   -0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.01  0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.   -0.   -0.    0.    0.   -0.   -0.    0.   -0.01  0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.01  0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.01  0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.01]]

nGC Level: 3
[[6.9995e-01 6.7688e-02 6.7688e-02 4.5166e-03 6.7688e-02 4.5166e-03 4.5166e-03 5.4932e-04 6.7688e-02 4.5166e-03 4.5166e-03 5.4932e-04 4.5166e-03 5.4932e-04 5.4932e-04 0.0000e+00]
 [6.7688e-02 6.9885e-01 4.6387e-03 5.9387e-02 4.6387e-03 5.9387e-02 6.7139e-04 1.1841e-02 4.6387e-03 5.9387e-02 6.7139e-04 1.1841e-02 6.7139e-04 1.1841e-02 0.0000e+00 3.8452e-03]
 [6.7688e-02 4.6387e-03 6.9885e-01 5.9387e-02 4.6387e-03 6.7139e-04 5.9387e-02 1.1841e-02 4.6387e-03 6.7139e-04 5.9387e-02 1.1841e-02 6.7139e-04 0.0000e+00 1.1841e-02 3.8452e-03]
 [4.5166e-03 5.9387e-02 5.9387e-02 4.7729e-01 6.7139e-04 1.9287e-02 1.9287e-02 1.2225e-01 6.7139e-04 1.9287e-02 1.9287e-02 1.2225e-01 0.0000e+00 1.0803e-02 1.0803e-02 5.4810e-02]
 [6.7688e-02 4.6387e-03 4.6387e-03 6.7139e-04 6.9885e-01 5.9387e-02 5.9387e-02 1.1841e-02 4.6387e-03 6.7139e-04 6.7139e-04 0.0000e+00 5.9387e-02 1.1841e-02 1.1841e-02 3.8452e-03]
 [4.5166e-03 5.9387e-02 6.7139e-04 1.9287e-02 5.9387e-02 4.7729e-01 1.9287e-02 1.2225e-01 6.7139e-04 1.9287e-02 0.0000e+00 1.0803e-02 1.9287e-02 1.2225e-01 1.0803e-02 5.4810e-02]
 [4.5166e-03 6.7139e-04 5.9387e-02 1.9287e-02 5.9387e-02 1.9287e-02 4.7729e-01 1.2225e-01 6.7139e-04 0.0000e+00 1.9287e-02 1.0803e-02 1.9287e-02 1.0803e-02 1.2225e-01 5.4810e-02]
 [5.4932e-04 1.1841e-02 1.1841e-02 1.2225e-01 1.1841e-02 1.2225e-01 1.2225e-01 2.3450e-01 0.0000e+00 1.0803e-02 1.0803e-02 6.7627e-02 1.0803e-02 6.7627e-02 6.7627e-02 1.2738e-01]
 [6.7688e-02 4.6387e-03 4.6387e-03 6.7139e-04 4.6387e-03 6.7139e-04 6.7139e-04 0.0000e+00 6.9885e-01 5.9387e-02 5.9387e-02 1.1841e-02 5.9387e-02 1.1841e-02 1.1841e-02 3.8452e-03]
 [4.5166e-03 5.9387e-02 6.7139e-04 1.9287e-02 6.7139e-04 1.9287e-02 0.0000e+00 1.0803e-02 5.9387e-02 4.7729e-01 1.9287e-02 1.2225e-01 1.9287e-02 1.2225e-01 1.0803e-02 5.4810e-02]
 [4.5166e-03 6.7139e-04 5.9387e-02 1.9287e-02 6.7139e-04 0.0000e+00 1.9287e-02 1.0803e-02 5.9387e-02 1.9287e-02 4.7729e-01 1.2225e-01 1.9287e-02 1.0803e-02 1.2225e-01 5.4810e-02]
 [5.4932e-04 1.1841e-02 1.1841e-02 1.2225e-01 0.0000e+00 1.0803e-02 1.0803e-02 6.7627e-02 1.1841e-02 1.2225e-01 1.2225e-01 2.3450e-01 1.0803e-02 6.7627e-02 6.7627e-02 1.2738e-01]
 [4.5166e-03 6.7139e-04 6.7139e-04 0.0000e+00 5.9387e-02 1.9287e-02 1.9287e-02 1.0803e-02 5.9387e-02 1.9287e-02 1.9287e-02 1.0803e-02 4.7729e-01 1.2225e-01 1.2225e-01 5.4810e-02]
 [5.4932e-04 1.1841e-02 0.0000e+00 1.0803e-02 1.1841e-02 1.2225e-01 1.0803e-02 6.7627e-02 1.1841e-02 1.2225e-01 1.0803e-02 6.7627e-02 1.2225e-01 2.3450e-01 6.7627e-02 1.2738e-01]
 [5.4932e-04 0.0000e+00 1.1841e-02 1.0803e-02 1.1841e-02 1.0803e-02 1.2225e-01 6.7627e-02 1.1841e-02 1.0803e-02 1.2225e-01 6.7627e-02 1.2225e-01 6.7627e-02 2.3450e-01 1.2738e-01]
 [0.0000e+00 3.8452e-03 3.8452e-03 5.4810e-02 3.8452e-03 5.4810e-02 5.4810e-02 1.2738e-01 3.8452e-03 5.4810e-02 5.4810e-02 1.2738e-01 5.4810e-02 1.2738e-01 1.2738e-01 1.4624e-01]]

nGC Level: 3
[[6.8018e-01 7.2083e-02 7.2083e-02 4.8828e-03 7.2083e-02 4.8828e-03 4.8828e-03 5.4932e-04 7.2083e-02 4.8828e-03 4.8828e-03 5.4932e-04 4.8828e-03 5.4932e-04 5.4932e-04 0.0000e+00]
 [7.2083e-02 6.7981e-01 5.0049e-03 6.3293e-02 5.0049e-03 6.3293e-02 6.7139e-04 1.2451e-02 5.0049e-03 6.3293e-02 6.7139e-04 1.2451e-02 6.7139e-04 1.2451e-02 0.0000e+00 3.8452e-03]
 [7.2083e-02 5.0049e-03 6.7981e-01 6.3293e-02 5.0049e-03 6.7139e-04 6.3293e-02 1.2451e-02 5.0049e-03 6.7139e-04 6.3293e-02 1.2451e-02 6.7139e-04 0.0000e+00 1.2451e-02 3.8452e-03]
 [4.8828e-03 6.3293e-02 6.3293e-02 4.6167e-01 6.7139e-04 1.9897e-02 1.9897e-02 1.2396e-01 6.7139e-04 1.9897e-02 1.9897e-02 1.2396e-01 0.0000e+00 1.0803e-02 1.0803e-02 5.6396e-02]
 [7.2083e-02 5.0049e-03 5.0049e-03 6.7139e-04 6.7981e-01 6.3293e-02 6.3293e-02 1.2451e-02 5.0049e-03 6.7139e-04 6.7139e-04 0.0000e+00 6.3293e-02 1.2451e-02 1.2451e-02 3.8452e-03]
 [4.8828e-03 6.3293e-02 6.7139e-04 1.9897e-02 6.3293e-02 4.6167e-01 1.9897e-02 1.2396e-01 6.7139e-04 1.9897e-02 0.0000e+00 1.0803e-02 1.9897e-02 1.2396e-01 1.0803e-02 5.6396e-02]
 [4.8828e-03 6.7139e-04 6.3293e-02 1.9897e-02 6.3293e-02 1.9897e-02 4.6167e-01 1.2396e-01 6.7139e-04 0.0000e+00 1.9897e-02 1.0803e-02 1.9897e-02 1.0803e-02 1.2396e-01 5.6396e-02]
 [5.4932e-04 1.2451e-02 1.2451e-02 1.2396e-01 1.2451e-02 1.2396e-01 1.2396e-01 2.2498e-01 0.0000e+00 1.0803e-02 1.0803e-02 6.9214e-02 1.0803e-02 6.9214e-02 6.9214e-02 1.2518e-01]
 [7.2083e-02 5.0049e-03 5.0049e-03 6.7139e-04 5.0049e-03 6.7139e-04 6.7139e-04 0.0000e+00 6.7981e-01 6.3293e-02 6.3293e-02 1.2451e-02 6.3293e-02 1.2451e-02 1.2451e-02 3.8452e-03]
 [4.8828e-03 6.3293e-02 6.7139e-04 1.9897e-02 6.7139e-04 1.9897e-02 0.0000e+00 1.0803e-02 6.3293e-02 4.6167e-01 1.9897e-02 1.2396e-01 1.9897e-02 1.2396e-01 1.0803e-02 5.6396e-02]
 [4.8828e-03 6.7139e-04 6.3293e-02 1.9897e-02 6.7139e-04 0.0000e+00 1.9897e-02 1.0803e-02 6.3293e-02 1.9897e-02 4.6167e-01 1.2396e-01 1.9897e-02 1.0803e-02 1.2396e-01 5.6396e-02]
 [5.4932e-04 1.2451e-02 1.2451e-02 1.2396e-01 0.0000e+00 1.0803e-02 1.0803e-02 6.9214e-02 1.2451e-02 1.2396e-01 1.2396e-01 2.2498e-01 1.0803e-02 6.9214e-02 6.9214e-02 1.2518e-01]
 [4.8828e-03 6.7139e-04 6.7139e-04 0.0000e+00 6.3293e-02 1.9897e-02 1.9897e-02 1.0803e-02 6.3293e-02 1.9897e-02 1.9897e-02 1.0803e-02 4.6167e-01 1.2396e-01 1.2396e-01 5.6396e-02]
 [5.4932e-04 1.2451e-02 0.0000e+00 1.0803e-02 1.2451e-02 1.2396e-01 1.0803e-02 6.9214e-02 1.2451e-02 1.2396e-01 1.0803e-02 6.9214e-02 1.2396e-01 2.2498e-01 6.9214e-02 1.2518e-01]
 [5.4932e-04 0.0000e+00 1.2451e-02 1.0803e-02 1.2451e-02 1.0803e-02 1.2396e-01 6.9214e-02 1.2451e-02 1.0803e-02 1.2396e-01 6.9214e-02 1.2396e-01 6.9214e-02 2.2498e-01 1.2518e-01]
 [0.0000e+00 3.8452e-03 3.8452e-03 5.6396e-02 3.8452e-03 5.6396e-02 5.6396e-02 1.2518e-01 3.8452e-03 5.6396e-02 5.6396e-02 1.2518e-01 5.6396e-02 1.2518e-01 1.2518e-01 1.4551e-01]]

nGC Level 3 error
[[-0.02  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.   -0.02  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.   -0.02  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.   -0.02  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.   -0.02  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.   -0.02  0.    0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.   -0.02  0.    0.    0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.   -0.01  0.    0.    0.    0.    0.    0.    0.   -0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.   -0.02  0.    0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.02  0.    0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.02  0.    0.    0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.01  0.    0.    0.   -0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.02  0.    0.    0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.01  0.   -0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.    0.   -0.01 -0.  ]
 [ 0.    0.    0.    0.    0.    0.    0.   -0.    0.    0.    0.   -0.    0.   -0.   -0.   -0.  ]]

'''